
import typing as t

from nr.util.generic import T
from ._base import BaseProxy


class StackableProxy(BaseProxy[T]):
  " Base class for proxies that can act as a stack and delegate to the item at the top of the stack. "

  def _empty(self) -> bool:
    raise NotImplementedError

  def _push(self, value: T) -> None:
    raise NotImplementedError

  def _pop(self) -> T:
    raise NotImplementedError


def empty(p: t.Union[StackableProxy[T], T]) -> bool:
  " Returns `True` if the stackable proxy is empty. "
  return t.cast(StackableProxy[T], p)._empty()


def push(p: t.Union[StackableProxy[T], T], value: T) -> None:
  " Push a value on a stackable proxy. "
  t.cast(StackableProxy[T], p)._push(value)


def pop(p: t.Union[StackableProxy[T], T]) -> T:
  " Pop a value from a stackable proxy. "
  return t.cast(StackableProxy[T], p)._pop()




import typing as t

from nr.util.generic import T


@t.overload
def coalesce(value: t.Optional[T], fallback: T) -> T: ...

@t.overload
def coalesce(value: t.Optional[T], *values: t.Optional[T]) -> t.Optional[T]: ...

@t.overload
def coalesce(value: t.Optional[T], *values: t.Optional[T], fallback: T) -> T: ...

def coalesce(value: t.Optional[T], *values: t.Optional[T], fallback: t.Optional[T] = None) -> t.Optional[T]:
  """
  Returns the first value that is not `None`. If a not-None fallback is specified, the function is guaranteed
  to return a not-Non value.
  """

  if value is not None:
    return value

  value = next((x for x in values if x is not None), None)
  if value is not None:
    return value

  return fallback




import typing as t

T = t.TypeVar('T')
T_co = t.TypeVar('T_co', covariant=True)
T_contra = t.TypeVar('T_contra', contravariant=True)

K = t.TypeVar('K')
K_co = t.TypeVar('K_co', covariant=True)
K_contra = t.TypeVar('K_contra', contravariant=True)

V = t.TypeVar('V')
V_co = t.TypeVar('V_co', covariant=True)
V_contra = t.TypeVar('V_contra', contravariant=True)

R = t.TypeVar('R')
R_co = t.TypeVar('R_co', covariant=True)
R_contra = t.TypeVar('R_contra', contravariant=True)

U = t.TypeVar('U')
U_co = t.TypeVar('U_co', covariant=True)
U_contra = t.TypeVar('U_contra', contravariant=True)

__all__ = [
  'T', 'T_co', 'T_contra',
  'K', 'K_co', 'K_contra',
  'V', 'V_co', 'V_contra',
  'R', 'R_co', 'R_contra',
  'U', 'U_co', 'U_contra',
]




import typing as t

from nr.util.generic import T, R

if t.TYPE_CHECKING:
  from nr.util import Stream


class Optional(t.Generic[T]):

  def __init__(self, value: t.Optional[T]) -> None:
    self._value = value

  def __bool__(self) -> bool:
    return self._value is not None

  def get(self) -> T:
    if self._value is None:
      raise ValueError('Optional value is None')
    return self._value

  def or_else(self, fallback: R) -> t.Union[T, R]:
    if self._value is None:
      return fallback
    return self._value

  def or_else_get(self, f: t.Callable[[], R]) -> t.Union[T, R]:
    if self._value is None:
      return f()
    return self._value

  def or_throw(self, exc: t.Union[Exception, t.Callable[[], Exception]]) -> T:
    if self._value is None:
      if callable(exc):
        raise exc()
      else:
        raise exc
    return self._value

  def map(self, f: t.Callable[[T], t.Optional[R]]) -> 'Optional[R]':
    if self._value is None:
      return Optional(None)
    return Optional(f(self._value))

  def flatmap(self, f: t.Callable[[T], t.Iterable[R]]) -> 'Stream[R]':
    return self.stream().flatmap(f)

  def stream(self) -> 'Stream[T]':
    from nr.util import Stream
    if self._value is None:
      return Stream()
    return Stream([self._value])




""" A set of helper functions that make it easy to attach arbitrary data to functions or classes. """

import types
import typing as t

from nr.util.generic import T

_ANNOTATION_MEMBER_NAME = '__nr_util_annotations__'

Annotateable = types.FunctionType | type | t.Any


def add_annotation(obj: Annotateable, annotation_base_type: t.Type[T], annotation: T, front: bool = False) -> None:
  """
  Registers an annotation in the given object, grouped under the specified base type. All annotations added to
  the object can be retrieved with #get_annotations() or #get_annotation(). If *front* is enabled, the annotation
  will be inserted at the front of the group instead of appended to the back.
  """

  if not hasattr(obj, _ANNOTATION_MEMBER_NAME):
    setattr(obj, _ANNOTATION_MEMBER_NAME, {})

  annotations: dict[type, list] = getattr(obj, _ANNOTATION_MEMBER_NAME)
  items = annotations.setdefault(annotation_base_type, [])
  items.insert(0, annotation) if front else items.append(annotation)


def get_annotations(obj: Annotateable, annotation_base_type: t.Type[T]) -> list[T]:
  """
  Returns all annotations registered on the given object in the group of the base type.
  """

  annotations: dict[type, list] = getattr(obj, _ANNOTATION_MEMBER_NAME, {})
  return list(annotations.get(annotation_base_type, []))


def get_annotation(obj: Annotateable, annotation_base_type: t.Type[T]) -> T | None:
  """
  Returns the first annotation registered to the given object in the group of the base type,
  or `None` if no annotations are registered for the given base type.
  """

  annotations = get_annotations(obj, annotation_base_type)
  return annotations[0] if annotations else None




"""
Provides classes that proxy all operations to a delegate object, allowing to implement behaviours at runtime where
the object contained in a "variable" can be interchanged or different based on the async context or thread.
"""

from ._base import get_name, get, BaseProxy
from ._proxy import bind, is_bound, proxy, Proxy
from ._stackable import empty, push, pop, StackableProxy
from ._threadlocal import threadlocal, ThreadLocalProxy
from ._contextlocal import contextlocal, ContextLocalProxy


__all__ = [
  'get_name', 'get', 'BaseProxy',
  'bind', 'is_bound', 'proxy', 'Proxy',
  'empty', 'push', 'pop', 'StackableProxy',
  'threadlocal', 'ThreadLocalProxy',
  'contextlocal', 'ContextLocalProxy',
]




import threading
import typing as t

from nr.util.generic import T
from ._stackable import StackableProxy


class ThreadLocalProxy(StackableProxy[T]):
  """
  A proxy that contains a thread-local stack of objects.
  """

  _local: threading.local
  _error_message: t.Optional[str]
  __name__: t.Optional[str]

  def __init__(self, name: t.Optional[str] = None, error_message: t.Optional[str] = None) -> None:
    """
    Create a new thread-local proxy. If an *error_message* is provided, it will be the
    message of the #RuntimeError that will be raised if the thread local is accessed
    without being initialized in the same thread first.
    """

    object.__setattr__(self, '_local', threading.local())
    object.__setattr__(self, '_error_message', error_message)
    object.__setattr__(self, '__name__', name)

  def _get_current_object(self) -> T:
    stack: t.List[T] = getattr(self._local, 'stack', [])
    if not stack:
      message = self._error_message or 'threadlocal {name} is not initialized in this thread'
      raise RuntimeError(message.format(name=self.__name__ or '<unnamed>'))
    return stack[-1]

  def _empty(self) -> bool:
    stack: t.List[T] = getattr(self._local, 'stack', [])
    return not stack

  def _push(self, value: T) -> None:
    assert value is not self, "cannot push threadlocal on itself"
    stack: t.Optional[t.List[T]] = getattr(self._local, 'stack', None)
    if stack is None:
      stack = []
      self._local.stack = stack
    stack.append(value)

  def _pop(self) -> T:
    stack: t.List[T] = getattr(self._local, 'stack', [])
    if not stack:
      name = self.__name__ or '<unnamed>'
      raise RuntimeError('there is no value to pop from threadlocal {}'.format(name))
    return stack.pop()


def threadlocal(name: t.Optional[str] = None, error_message: t.Optional[str] = None) -> t.Any:
  " Factory function to create a #ThreadLocalProxy object, but typed to return any for convenience. "
  return ThreadLocalProxy(name, error_message)




""" General purpose utility library. """

__author__ = 'Niklas Rosenstein <rosensteinniklas@gmail.com>'
__version__ = "0.1.0"

from ._coalesce import coalesce
from ._optional import Optional
from ._refreshable import Refreshable
from ._stream import Stream

__all__ = ['coalesce', 'Optional', 'Refreshable', 'Stream']




import logging
import threading
import typing as t

from nr.util.generic import T, R

logger = logging.getLogger(__name__)
Subscriber = t.Callable[['Refreshable[T]'], None]


class Refreshable(t.Generic[T]):

  def __init__(self, initial: T) -> None:
    self._lock = threading.Lock()
    self._current = initial
    self._subscribers: t.List[Subscriber] = []

  def get(self) -> T:
    with self._lock:
      return self._current

  def update(self, value: T) -> None:
    with self._lock:
      self._current = value
      subscribers = self._subscribers[:]
    for subscriber in subscribers:
      try:
        subscriber(self)
      except:
        logger.exception('Error in subscriber')

  def subscribe(self, subscriber: Subscriber) -> None:
    with self._lock:
      self._subscribers.append(subscriber)
    try:
      subscriber(self)
    except:
      logger.exception('Error in subscriber')

  def map(self, mapper: t.Callable[[T], R]) -> 'Refreshable[R]':
    """
    Map the value of the refreshable to a new refreshable that automatically gets updated when the
    parent is updated. Be aware that this method should be used sparingly as it registers a new
    subscriber to this refreshable that will never be disposed of again.
    """

    child = Refreshable(mapper(self.get()))
    def refresh(_parent):
      child.update(mapper(self.get()))
    self.subscribe(refresh)
    return child




import typing as t

from nr.util.function import Supplier
from nr.util.generic import T
from ._base import BaseProxy


class Proxy(BaseProxy[T]):
  """
  Wraps an object returned by a callable. Every time the proxy is accessed, the callable is invoked and the access is
  redirected to the proxied object.

  > __Note__: For some use cases, using a proxy can lead to problems when `isinstance()` or checks are used. For
  > example if you pass a proxy to a function that accepts an iterable, the instancecheck with
  > #collections.abc.Iterable will always return `True` as the #Proxy class implements the `__iter__()` method. If
  > you are stuck with this problem, use the #make_cls() function to create a new class excluding the `__iter__()`
  > method.
  """

  _func: Supplier[T]
  _lazy: bool
  _cache: t.Optional[T] = None

  def __init__(self, func: Supplier[T] = None, name: t.Optional[str] = None, lazy: bool = False) -> None:
    """
    Create a new proxy object.

    # Arguments
    func (callable): The function that will be called to retrieve the object represented by the proxy when demanded.
      This will be called again and again for every operation performed on the proxy, unless #lazy is enabled. If no
      func is set, the proxy is unbound. The proxy can later be bound using the #bind() function.
    name (str): The name of the proxy. Can make debugging easier if the proxy is unbound.
    lazy (bool): Call the *func* function only once on first use, then cache the result.
    """

    object.__setattr__(self, "_func", func)
    object.__setattr__(self, "_lazy", lazy)
    object.__setattr__(self, "_cache", None)
    object.__setattr__(self, "__name__", name)

  def _set(self, value: T) -> None:
    object.__setattr__(self, "_func", None)
    object.__setattr__(self, "_lazy", True)
    object.__setattr__(self, "_cache", value)

  # proxy_base

  def _get_current_object(self) -> T:
    if self._lazy:
      if self._cache is None:
        if self._func is None:
          raise RuntimeError('unbound proxy')
        object.__setattr__(self, "_cache", self._func())
      return t.cast(T, self._cache)
    elif self._func is None:
      raise RuntimeError('unbound proxy')
    else:
      return self._func()

  def _is_bound(self) -> bool:
    return self._func is not None

  def _bind(self, func: t.Optional[Supplier[T]]) -> None:
    self._func = func  # type: ignore


def set_value(p: t.Union[Proxy[T], T], value: T) -> None:
  " Permanently override the value of a #proxy. This will turn the proxy to a lazy proxy if it is not already that. "

  t.cast(Proxy[T], p)._set(value)


def bind(p: t.Union[Proxy[T], T], func: t.Optional[Supplier[T]]) -> None:
  """
  (Re-) bind the function for a proxy. The *func* will be called in the future when the current value of the
  proxy is required.
  """

  t.cast(Proxy[T], p)._bind(func)


def is_bound(p: t.Union[Proxy[T], T]) -> bool:
  return t.cast(Proxy[T], p)._is_bound()


def proxy(func: Supplier[T], name: t.Optional[str] = None, lazy: bool = False) -> T:
  """
  Create a new proxy from the given supplier. The function is typed to return an instance of the generic type
  *T*, but will in fact return a #Proxy object.
  """

  return t.cast(T, Proxy(func, name, lazy))





import collections
import functools
import itertools
import typing as t

from nr.util.generic import T, T_co, U, R
from nr.util.singleton import NotSet

if t.TYPE_CHECKING:
  from nr.util import Optional

Aggregator = t.Callable[[T, U], T]
Collector = t.Callable[[t.Iterable[T]], R]


class Stream(t.Generic[T_co], t.Iterable[T_co]):
  """
  A stream is an iterable with utility methods to transform it.
  """

  def __init__(self, iterable: t.Optional[t.Iterable[T_co]] = None) -> None:
    if iterable is None:
      iterable = ()
    self._it = iter(iterable)
    self._original: t.Optional[t.Iterable[T_co]] = iterable

  def __iter__(self) -> 'Stream[T_co]':
    return self

  def __next__(self) -> T_co:
    self._original = None
    return next(self._it)

  @t.overload
  def __getitem__(self, val: slice) -> 'Stream[T_co]': ...

  @t.overload
  def __getitem__(self, val: int) -> T_co: ...

  def __getitem__(self, val):
    if isinstance(val, slice):
      return self.slice(val.start, val.stop, val.step)
    elif isinstance(val, int):
      if val >= 0:
        for index, value in enumerate(self):
          if index == val:
            return value
        raise IndexError('Stream has no element at position {}'.format(val))
      else:
        queue = collections.deque(self, maxlen=abs(val))
        if len(queue) < abs(val):
          raise IndexError('Stream has no element at position {}'.format(val))
        return queue[0]
    else:
      raise TypeError('{} object is only subscriptable with slices'.format(type(self).__name__))

  def next(self) -> T_co:
    return next(self._it)

  def append(self, *its: t.Iterable[T_co]) -> 'Stream[T_co]':
    return Stream(itertools.chain(self._it, *its))

  @t.overload
  def batch(self, n: int) -> 'Stream[t.List[T_co]]': ...

  @t.overload
  def batch(self, n: int, collector: Collector[T_co, R]) -> 'Stream[R]': ...

  def batch(self, n, collector=None):
    """
    Convert the stream into a stream of batches of size *n*, where each element of the stream
    contains the result of the *collector* after passing up to *n* elements of the original
    stream into it.
    """

    iterable = iter(self._it)
    if collector is None:
      collector = list

    def take(first):
      yield first
      count = 1
      while count < n:
        try:
          yield next(iterable)
        except StopIteration:
          break
        count += 1

    def generate_batches():
      while True:
        try:
          first = next(iterable)
        except StopIteration:
          break
        yield collector(take(first))

    return Stream(generate_batches())

  def bipartition(self, predicate: t.Callable[[T_co], bool]) -> 't.Tuple[Stream[T_co], Stream[T_co]]':
    """
    Use a predicate to partition items into false and true entries.
    Returns a tuple of two streams with the first containing all elements
    for which *pred* returned #False and the other containing all elements
    where *pred* returned #True.
    """

    t1, t2 = itertools.tee(self._it)
    return Stream(itertools.filterfalse(predicate, t1)), Stream(filter(predicate, t2))

  def call(self: 'Stream[t.Callable[..., R]]', *a: t.Any, **kw: t.Any) -> 'Stream[R]':
    """
    Calls every item in *iterable* with the specified arguments.
    """

    return Stream(x(*a, **kw) for x in self._it)

  @t.overload
  def collect(self) -> t.List[T_co]: ...

  @t.overload
  def collect(self, collector: Collector[T_co, R]) -> R: ...

  def collect(self, collector=list):
    """
    Collects the stream into a collection.
    """

    if isinstance(collector, type) and isinstance(self._original, collector):
      # NOTE(NiklasRosenstein): This is an optimization to retrieve the original underlying
      #   collection if the stream has not been advanced yet.
      return self._original

    return collector(self._it)

  def count(self) -> int:
    """
    Returns the number of items in the stream. This fully consumes the stream.
    """

    count = 0
    while True:
      try:
       next(self._it)
      except StopIteration:
        break
      count += 1
    return count

  @t.overload
  def concat(s: "Stream[str]") -> "Stream[str]":
    ...

  @t.overload
  def concat(s: "Stream[t.Iterable[T]]") -> "Stream[T]":
    ...

  def concat(self):
    """
    Concatenate all values in the stream into a single stream of values.
    """

    def generator():
      for it in self:
        for element in it:
          yield element
    return Stream(generator())

  def consume(self, n: t.Optional[int] = None) -> 'Stream[T_co]':
    """
    Consume the contents of the stream, up to *n* elements if the argument is specified.
    """

    if n is not None:
      for _ in range(n):
        try:
          next(self._it)
        except StopIteration:
          break
    else:
      while True:
        try:
          next(self._it)
        except StopIteration:
          break
    return self

  def distinct(self,
    key: t.Optional[t.Callable[[T_co], t.Any]] = None,
    skip: t.Union[t.MutableSet[T_co], t.MutableSequence[T_co], None] = None,
  ) -> 'Stream[T_co]':
    """
    Yields unique items from *iterable* whilst preserving the original order. If *skip* is
    specified, it must be a set or sequence of items to skip in the first place (ie. items to
    exclude from the returned stream). The specified set/sequence is modified in-place. Using a
    set is highly recommended for performance purposes.
    """

    if key is None:
      key_func = lambda x: x
    else:
      key_func = key

    def generator() -> t.Generator[T_co, None, None]:
      seen = set() if skip is None else skip
      mark_visited = seen.add if isinstance(seen, t.MutableSet) else seen.append
      check_visited = seen.__contains__
      for item in self._it:
        key_val = key_func(item)
        if not check_visited(key_val):
          mark_visited(key_val)
          yield item

    return Stream(generator())

  def dropwhile(self, predicate: t.Callable[[T_co], bool]) -> 'Stream[T_co]':
    return Stream(itertools.dropwhile(predicate, self._it))

  def dropnone(self: 'Stream[t.Optional[T_co]]') -> 'Stream[T_co]':
    return Stream(x for x in self._it if x is not None)

  def filter(self, predicate: t.Callable[[T_co], bool]) -> 'Stream[T_co]':
    """
    Agnostic to Python's built-in `filter()` function.
    """

    return Stream(x for x in self._it if predicate(x))

  def first(self) -> t.Optional[T_co]:
    """
    Returns the first element of the stream, or `None`.
    """

    try:
      return self.next()
    except StopIteration:
      return None

  def firstopt(self) -> 'Optional[T_co]':
    """
    Returns the first element of the stream as an `Optional`.
    """

    from nr.util import Optional
    return Optional(self.first())

  def flatmap(self, func: t.Callable[[T_co], t.Iterable[R]]) -> 'Stream[R]':
    """
    Same as #map() but flattens the result.
    """

    def generator() -> t.Generator[R, None, None]:
      for x in self._it:
        for y in func(x):
          yield y
    return Stream(generator())

  @t.overload
  def groupby(self, key: t.Callable[[T_co], R]) -> 'Stream[t.Tuple[R, t.Iterable[T_co]]]': ...

  @t.overload
  def groupby(self, key: t.Callable[[T_co], R], collector: t.Callable[[t.Iterable[T_co]], U]) -> 'Stream[t.Tuple[R, U]]': ...

  def groupby(self, key: t.Callable[[T_co], U], collector: t.Optional[Collector[T_co, R]] = None):
    if collector is None:
      return Stream(itertools.groupby(self._it, key))
    else:
      def generator():
        assert collector is not None
        g: t.Iterable[T_co]
        for k, g in itertools.groupby(self._it, key):
          yield k, collector(g)
      return Stream(generator())

  def map(self, func: t.Callable[[T_co], R]) -> 'Stream[R]':
    """
    Agnostic to Python's built-in `map()` function.
    """

    return Stream(func(x) for x in self._it)

  def of_type(self, type: t.Type[T_co]) -> 'Stream[T_co]':
    """
    Filters using #isinstance().
    """

    return Stream(x for x in self._it if isinstance(x, type))

  @t.overload
  def reduce(self, aggregator: Aggregator[T_co, T_co]) -> T_co: ...

  @t.overload
  def reduce(self, aggregator: Aggregator[R, T_co], initial: R) -> R: ...

  def reduce(self, aggregator, initial=NotSet.Value):
    if initial is NotSet.Value:
      return functools.reduce(aggregator, self._it)
    else:
      return functools.reduce(aggregator, self._it, initial)

  @t.overload
  def slice(self, stop: int) -> 'Stream[T_co]': ...

  @t.overload
  def slice(self, start: int, stop: int, step: int = 1) -> 'Stream[T_co]': ...

  def slice(self, start, stop=None, step=None):
    return Stream(itertools.islice(self._it, start, stop, step))

  def sortby(self, by: t.Union[str, t.Callable[[T_co], t.Any]], reverse: bool = False) -> 'Stream[T_co]':
    """
    Creates a new sorted stream. Internally the #sorted() built-in function is used so a new list
    will be created temporarily.

    # Parameters
    by (str, callable): Specify by which dimension to sort the stream. If a string is specified,
      it will be used to retrieve a key or attribute from the values in the stream. In the case of
      a callable, it will be used directly as the `key` argument to #sorted().
    """

    if isinstance(by, str):
      lookup_attr = by
      def by(item):
        if isinstance(item, t.Mapping):
          return item[lookup_attr]
        else:
          return getattr(item, lookup_attr)

    by = t.cast(t.Callable[[T_co], t.Any], by)
    return Stream(sorted(self._it, key=by, reverse=reverse))

  def sort(self: 'Stream[T]', reverse: bool = False) -> 'Stream[T]':
    return self.sortby(lambda x: x, reverse)

  def takewhile(self, predicate: t.Callable[[T_co], bool]) -> 'Stream[T_co]':
    return Stream(itertools.takewhile(predicate, self._it))




import copy
import typing as t

from nr.util.generic import T


class BaseProxy(t.Generic[T]):
  " Base class for proxy objects, which implements all method delegation to #_get_current_object(). "

  __is_proxy__ = True

  def _get_current_object(self) -> t.Any:
    raise NotImplementedError

  # forward special function calls and attribute access

  @property
  def __dict__(self):
    try:
      return self._get_current_object().__dict__
    except RuntimeError:
      raise AttributeError("__dict__")

  def __repr__(self):
    try:
      obj = self._get_current_object()
    except RuntimeError:
      return "<unresolvable {}.{} ({})>".format(
        type(self).__module__,
        type(self).__name__,
        get_name(self) or '<unnamed>')
    return repr(obj)

  def __bool__(self):
    try:
      return bool(self._get_current_object())
    except RuntimeError:
      return False

  def __dir__(self):
    try:
      return dir(self._get_current_object())
    except RuntimeError:
      return []

  def __getattr__(self, name):
    if name == '__members__':
      return dir(self._get_current_object())
    return getattr(self._get_current_object(), name)

  def __setitem__(self, key, value):
    self._get_current_object()[key] = value

  def __delitem__(self, key):
    del self._get_current_object()[key]

  def __call__(self, *a, **kw):
    return self._get_current_object()(*a, **kw)

  def __setattr__(self, name, value):
    if name == '__orig_class__':
      # Support for generic instantiation of the proxy type.
      object.__setattr__(self, name, value)
    else:
      setattr(self._get_current_object(), name, value)

  __delattr__ = lambda x, n: delattr(x._get_current_object(), n)  # type: ignore
  __str__ = lambda x: str(x._get_current_object())  # type: ignore
  __lt__ = lambda x, o: x._get_current_object() < o
  __le__ = lambda x, o: x._get_current_object() <= o
  __eq__ = lambda x, o: x._get_current_object() == o  # type: ignore
  __ne__ = lambda x, o: x._get_current_object() != o  # type: ignore
  __gt__ = lambda x, o: x._get_current_object() > o
  __ge__ = lambda x, o: x._get_current_object() >= o
  __hash__ = lambda x: hash(x._get_current_object())  # type: ignore
  __len__ = lambda x: len(x._get_current_object())
  __getitem__ = lambda x, i: x._get_current_object()[i]
  __iter__ = lambda x: iter(x._get_current_object())
  __contains__ = lambda x, i: i in x._get_current_object()
  __add__ = lambda x, o: x._get_current_object() + o
  __sub__ = lambda x, o: x._get_current_object() - o
  __mul__ = lambda x, o: x._get_current_object() * o
  __floordiv__ = lambda x, o: x._get_current_object() // o
  __mod__ = lambda x, o: x._get_current_object() % o
  __divmod__ = lambda x, o: x._get_current_object().__divmod__(o)
  __pow__ = lambda x, o: x._get_current_object() ** o
  __lshift__ = lambda x, o: x._get_current_object() << o
  __rshift__ = lambda x, o: x._get_current_object() >> o
  __and__ = lambda x, o: x._get_current_object() & o
  __xor__ = lambda x, o: x._get_current_object() ^ o
  __or__ = lambda x, o: x._get_current_object() | o
  __div__ = lambda x, o: x._get_current_object().__div__(o)
  __truediv__ = lambda x, o: x._get_current_object().__truediv__(o)
  __neg__ = lambda x: -(x._get_current_object())
  __pos__ = lambda x: +(x._get_current_object())
  __abs__ = lambda x: abs(x._get_current_object())
  __invert__ = lambda x: ~(x._get_current_object())
  __complex__ = lambda x: complex(x._get_current_object())
  __int__ = lambda x: int(x._get_current_object())
  __float__ = lambda x: float(x._get_current_object())
  __oct__ = lambda x: oct(x._get_current_object())
  __hex__ = lambda x: hex(x._get_current_object())
  __index__ = lambda x: x._get_current_object().__index__()
  __coerce__ = lambda x, o: x._get_current_object().__coerce__(x, o)
  __enter__ = lambda x: x._get_current_object().__enter__()
  __exit__ = lambda x, *a, **kw: x._get_current_object().__exit__(*a, **kw)
  __radd__ = lambda x, o: o + x._get_current_object()
  __rsub__ = lambda x, o: o - x._get_current_object()
  __rmul__ = lambda x, o: o * x._get_current_object()
  __rdiv__ = lambda x, o: o / x._get_current_object()
  __rtruediv__ = __rdiv__
  __rfloordiv__ = lambda x, o: o // x._get_current_object()
  __rmod__ = lambda x, o: o % x._get_current_object()
  __rdivmod__ = lambda x, o: x._get_current_object().__rdivmod__(o)
  __copy__ = lambda x: copy.copy(x._get_current_object())
  __deepcopy__ = lambda x, memo: copy.deepcopy(x._get_current_object(), memo)
  __class__ = property(lambda x: type(x._get_current_object()))  # type: ignore


def get(p: t.Union[BaseProxy[T], T]) -> T:
  " Dereference the proxy to get it's current value. "
  return t.cast(BaseProxy[T], p)._get_current_object()


def get_name(p: t.Union[BaseProxy[T], T]) -> t.Optional[str]:
  " Returns the name of a proxy object, if one was specified on construction. "
  try:
    return t.cast(str, object.__getattribute__(p, '__name__'))
  except AttributeError:
    return None
